@@ -1,7 +1,10 @@
+from abc import ABC, abstractmethod
 from copy import deepcopy
+from functools import reduce
+import math
+from numbers import Number
 from sortedcontainers import SortedDict
 from typing import List, Tuple, Optional, Union, Dict, Callable
-from numbers import Number
 
 
 def ensure_sorteddict_of_keyframes(curve: 'Curve',default_interpolation:Union[str,Callable]='previous') -> SortedDict:
@@ -12,21 +15,50 @@ def ensure_sorteddict_of_keyframes(curve: 'Curve',default_interpolation:Union[st
     - If it is a tuple, it is assumed to be in the format ((k0,v0), (k1,v1), ...).
     """
     if isinstance(curve, SortedDict):
-        outv = curve
+        sorteddict = curve
     elif isinstance(curve, dict):
-        outv = SortedDict(curve)
+        sorteddict = SortedDict(curve)
     elif isinstance(curve, Number):
-        outv = SortedDict({0:Keyframe(t=0,value=curve, interpolation_method=default_interpolation)})
-    elif isinstance(curve, tuple):
-        outv = SortedDict({k:Keyframe(t=k,value=v, interpolation_method=default_interpolation) for k,v in curve})
+        sorteddict = SortedDict({0:Keyframe(t=0,value=curve, interpolation_method=default_interpolation)})
+    elif (isinstance(curve, list) or isinstance(curve, tuple)):
+        d_ = {}
+        for item in curve:
+            if isinstance(item, Keyframe):
+                d_[item.t] = item
+            else:
+                k,v = item
+                d_[k] = v
+        sorteddict = SortedDict(d_)
     else:
         raise NotImplementedError
-    if 0 not in outv:
-        outv[0] = 0
-    for k, v in list(outv.items()):
-        if not isinstance(v, Keyframe):
-            outv[k] = Keyframe(t=k,value=v, interpolation_method=default_interpolation)
-    return outv
+
+    d_ = {}
+    implied_interpolation = default_interpolation
+    if 0 not in sorteddict:
+        d_[0] = Keyframe(t=0,value=0, interpolation_method=implied_interpolation)
+    for k,v in sorteddict.items():
+        if isinstance(v, Keyframe):
+            implied_interpolation = v.interpolation_method
+            d_[k] = v
+        elif isinstance(v, dict):
+            kf = Keyframe(**v)
+            if 'interpolation_method' not in v:
+                kf.interpolation_method = implied_interpolation
+            implied_interpolation = kf.interpolation_method
+            if k != kf.t:
+                kf.t = k
+            d_[k] = kf
+        elif isinstance(v, list) or isinstance(v, tuple):
+            kf = Keyframe(*v)
+            if len(v) < 3:
+                kf.interpolation_method = implied_interpolation
+            implied_interpolation = kf.interpolation_method
+            d_[k] = kf
+        elif isinstance(v, Number):
+            d_[k] = Keyframe(t=k,value=v, interpolation_method=implied_interpolation)
+        else:
+            raise NotImplementedError
+    return SortedDict(d_)
 
 
 def bisect_left_keyframe(k: Number, curve:'Curve') -> 'Keyframe':
@@ -71,6 +103,7 @@ def bisect_right_value(k: Number, curve:'Curve') -> 'Keyframe':
     return kf.value
 
 # NB: interp1d only supports INTERPOLATION, not extrapolation.
+# ... increasingly unsure how I feel about this, as implemented at least.
 def scipy_interp(k:Number, curve:'Curve', kind:str, **kargs) -> Callable:
     """
     Wraps scipy.interpolate.interp1d for use with Curve objects.
@@ -83,10 +116,37 @@ def scipy_interp(k:Number, curve:'Curve', kind:str, **kargs) -> Callable:
     f = interp1d(x=xs, y=ys, kind=kind, **kargs)
     return f
 
+def sin2(t:Number) -> Number:
+    return (math.sin(t * math.pi / 2)) ** 2
+
+# to do: turn this into a decorator in dmarx/Keyframed
+def eased_lerp(k:Number, curve:'Curve', ease:Callable=sin2) -> Number:
+    left = bisect_left_keyframe(k, curve)
+    right = bisect_right_keyframe(k, curve)
+    xs = [left.t, right.t]
+    ys = [left.value, right.value]
+
+    span = xs[1]-xs[0]
+    t = (k-xs[0]) / span
+    t_new = ease(t)
+    return ys[1] * t_new + ys[0] * (1-t_new)
+
+def linear(k, curve):
+    left = bisect_left_keyframe(k, curve)
+    right = bisect_right_keyframe(k, curve)
+    x0, x1 = left.t, right.t
+    y0, y1 = left.value, right.value
+    d = x1-x0
+    t = (x1-k)/d
+    outv =  t*y0 + (1-t)*y1
+    return outv
+
 INTERPOLATORS={
     None:bisect_left_value,
     'previous':bisect_left_value,
     'next':bisect_right_value,
+    'eased_lerp':eased_lerp,
+    'linear':linear,
 }
 
 def register_interpolation_method(name:str, f:Callable):
@@ -228,19 +288,66 @@ class EaseOut(EasingFunction):
             k_prev = k
         else:
             return self.curve.keyframes[-1]
-    # def __call__(self,k:Number) -> Number:
-    #     if not self.use_easing(k):
-    #         return k
-    #     span = self.end_t - self.start_t
-    #     #t = (k-self.start_t) / span
-    #     t = (self.end_t-k) / span
-    #     t_new = self.f(t)
-    #     k_new = self.start_t + t_new*span
-    #     return k_new
 
-class Curve:
+
+class CurveBase(ABC):
+    def copy(self):
+        return deepcopy(self)
+
+    @property
+    @abstractmethod
+    def keyframes(self):
+        pass
+
+    @property
+    @abstractmethod
+    def values(self):
+        pass
+    
+    @property
+    @abstractmethod
+    def duration(self):
+        pass 
+
+    @abstractmethod
+    def __getitem__(self):
+        pass
+
+    def plot(self, n:int=None, xs:list=None, eps:float=1e-9, *args, **kargs):
+        """
+        Arguments
+            n (int): (Optional) Number of points to plot, plots range [0,n-1]. If not specified, n=self.duration.
+            eps (float): (Optional) value to be subtracted from keyframe to produce additional points for plotting.
+                Plotting these additional values is important for e.g. visualizing step function behavior.
+        """
+        try:
+            import matplotlib.pyplot as plt
+        except ImportError:
+            raise ImportError("Please install matplotlib to use Curve.plot()")
+        if xs is None:
+            if n is None:
+                n = self.duration + 1
+            xs_base = list(range(int(n))) + list(self.keyframes)
+            xs = set()
+            for x in xs_base:
+                if (x>0) and (eps is not None) and (eps > 0):
+                    xs.add(x-eps)
+                xs.add(x)
+        xs = list(set(xs))
+        xs.sort()
+        
+        ys = [self[x] for x in xs]
+        if kargs.get('label') is None:
+            kargs['label']=self.label
+        plt.plot(xs, ys, *args, **kargs)
+        kfx = self.keyframes
+        kfy = [self[x] for x in kfx]
+        plt.scatter(kfx, kfy)
+
+
+class Curve(CurveBase):
     """
-    Represents a curve as a sorted dictionary of Keyframes.
+    Represents a curve as a sorted dictionary of Keyframes. Default interpolation produces a step function.
 
     Attributes:
         ease_in (str): The method used for easing in to the curve.
@@ -264,6 +371,7 @@ class Curve:
         ease_out:Union[EaseOut, Callable] = None,
         loop: bool = False,
         duration:Optional[float]=None,
+        label:str=None,
     ):
         """
         Initializes a curve from a dictionary or another curve.
@@ -300,7 +408,8 @@ class Curve:
         self.ease_out=ease_out
         self.loop=loop
         self._duration=duration
-        
+        self.label=label
+
     @property
     def keyframes(self):
         return self._data.keys()
@@ -310,10 +419,12 @@ class Curve:
         # not a fan of this
         return [kf.value for kf in self._data.values()]
 
-    def __len__(self):
+    @property
+    def duration(self):
         if self._duration:
             return self._duration
-        return max(self.keyframes)+1
+        #return max(self.keyframes)+1
+        return max(self.keyframes)
 
     def __getitem__(self, k:Number) -> Number:
         """
@@ -321,13 +432,13 @@ class Curve:
         but indexing into this class should always return a number (Keyframe.value)
         """
         if self.loop and k >= max(self.keyframes):
-            k %= len(self)
+            k %= (self.duration + 1)
         if k in self._data.keys():
             outv = self._data[k]
             if isinstance(outv, Keyframe):
                 outv = outv.value
             return outv
-        #if k > (len(self)-1):
+        #if k > self.duration:
         #    return 0
 
         left_value = bisect_left_keyframe(k, self)
@@ -371,50 +482,61 @@ class Curve:
         d_ = {k:self[k] for k in self.keyframes}
         return f"Curve({d_}"
 
-    ##########################
-    def copy(self):
-        return deepcopy(self)
-
-    def __add__(self, other) -> 'Curve':
-        if isinstance(other, type(self)):
+    def __add__(self, other) -> CurveBase:
+        if isinstance(other, CurveBase):
             return self.__add_curves__(other)
         outv = self.copy()
         for k in self.keyframes:
             outv[k]+=other
         return outv
 
-    def __add_curves__(self, other) -> 'Curve':
-        outv = self.copy()
-        other = other.copy()
+    def __to_labeled(self, other) -> dict:
+        self_label = self.label
+        if self_label is None:
+            self_label = 'this'
+        other_label = other.label
+        if other_label is None:
+            other_label = 'that'
+        if self_label == other_label:
+            other_label += '1'
+        return {self_label:self, other_label:other}
 
-        # step 1. lift current keyframes by what `other` evaluates to at that point
-        for k in outv.keyframes:
-            outv._data[k].value = self[k] + other[k]
+    def __add_curves__(self, other) -> 'Composition':
+        params = self.__to_labeled(other)
+        new_label = '+'.join(params.keys())
+        return Composition(parameters=params, label=new_label, reduction='add')
 
-        # step 2. perform the converse operation on `other` using what `self` evaluates to at its keyframes
-        for k in other.keyframes:
-            other[k] += self[k]
-
-        # step 3. for any keys which `other` has but `self` does not, pop that item from `other` and insert it into `outv`.
-        for k, kf in other._data.items():
-            if k not in outv.keyframes:
-                outv._data[k] = kf
-        
-        return outv
-
-    def __mul__(self, other) -> 'Curve':
+    def __mul__(self, other) -> CurveBase:
+        if isinstance(other, CurveBase):
+            return self.__mul_curves__(other)
         outv = self.copy()
         for i, k in enumerate(self.keyframes):
             kf = outv._data[k]
             kf.value = kf.value * other
             outv[k]=kf
         return outv
+    
+    def __mul_curves__(self, other) -> 'Composition':
+        params = self.__to_labeled(other)
+        pg = ParameterGroup(params)
+        new_label = '*'.join(params.keys())
+        return Composition(parameters=pg, label=new_label, reduction='multiply')
+
     def __rmul__(self, other) -> 'Curve':
         return self*other
 
 
+def SmoothCurve(*args, **kargs):
+    """
+    Thin wrapper around the Curve class that uses an 'eased_lerp' for `default_interpolation` to produce a smooth curve
+    instead of a step function. In the future, the interpolation function on this class may be modified to use a different
+    smoothing interpolator.
+    """
+    return Curve(*args, default_interpolation='eased_lerp', **kargs)
+
+
 # i'd kind of like this to inherit from dict.
-class ParameterGroup:
+class ParameterGroup(CurveBase):
     """
     The ParameterGroup class wraps a collection of named parameters to facilitate manipulating them as a unit.
     Indexing into a ParameterGroup with a frame index returns a dict giving the values of the parameters at that time.
@@ -422,16 +544,22 @@ class ParameterGroup:
     """
     def __init__(
         self,
-        parameters:Dict[str, Curve],
+        parameters:Union[Dict[str, Curve],'ParameterGroup'],
         weight:Optional[Union[Curve,Number]]=1
     ):
+        if isinstance(parameters, ParameterGroup):
+            pg = parameters
+            self.parameters = pg.parameters
+            self.weight = pg.weight
+            return
         if not isinstance(weight, Curve):
             weight = Curve(weight)
         self.weight = weight
         self.parameters={}
         for name, v in parameters.items():
-            if isinstance(v, Number):
+            if not isinstance(v, CurveBase):
                 v = Curve(v)
+            v.label = name
             self.parameters[name] = v
 
     def __getitem__(self, k) -> dict:
@@ -457,3 +585,71 @@ class ParameterGroup:
 
     def __rmul__(self, other) -> 'ParameterGroup':
         return self*other
+
+    @property
+    def duration(self):
+            return max(curve.duration for curve in self.parameters.values())
+
+    def plot(self):
+        n = self.duration + 1
+        for curve in self.parameters.values():
+            curve.plot(n=n)
+
+    @property
+    def keyframes(self):
+        kfs = set()
+        for curve in self.parameters.values():
+            kfs.update(curve.keyframes)
+        kfs = list(kfs) 
+        kfs.sort()
+        return kfs
+
+    @property
+    def values(self):
+        return [self[k] for k in self.keyframes]
+
+
+REDUCTIONS = {
+    'add':lambda x,y:x+y,
+    'multiply':lambda x,y:x*y,
+    'subtract':lambda x,y:x-y,
+    'divide':lambda x,y:x/y,
+}
+
+
+class Composition(ParameterGroup):
+    """
+    Synthesizes a new curve by performing a reduction operation over two or more
+    other curves. The value for a given keyframe k is computed by evaluating the
+    input curves at k, then performing the reduction operation over the resultant values.
+    The input curves are assumed to be passed by reference, so modifications to the
+    input Curve objects will propogate to compositions of those curves.
+
+    Arguments
+      parameters (ParameterGroup): The curves to be composed, encapsulated in a ParameterGroup
+      reduction (Callable): A function that defines the combination operator.
+      label (str): Optional label, e.g. for plotting. If not provided, one will be inferred from the parameters.
+    """
+    def __init__(
+        self,
+        parameters:Union[Dict[str, Curve],'ParameterGroup'],
+        weight:Optional[Union[Curve,Number]]=1,
+        reduction:str=None,
+        label:str=None,
+    ):
+        super().__init__(parameters=parameters, weight=weight)
+        self.reduction = reduction
+        self._label=label
+    def __getitem__(self, k):
+        f = REDUCTIONS.get(self.reduction)
+        vals = super().__getitem__(k).values()
+        outv = reduce(f, vals)
+        return outv
+    @property
+    def _default_label(self):
+        return ''.join([f"({k})" for k in self.parameters.keys()])
+    @property
+    def label(self):
+        if self._label is not None:
+            return self._label
+        return self._default_label
